from aiohttp import web
import json
import asyncio
import bcrypt

from logging import getLogger
_log = getLogger(__name__)


_instance = None


def get_websockets():
    global _instance
    return _instance


def setup_websockets(app=None, db=None, options=None):
    global _instance
    if _instance is not None:
        del _instance
    _instance = Websockets(app=app, db=db, options=options)
    return _instance


class Websockets():
    def __init__(self, app=None, db=None, options=None):
        global _instance
        if _instance is None:
            _instance = self

        self._options = options
        self._db = db

        self._ws_list = []
        self._subscriptions = {}

        app.router.add_get(self._options.websocket_path,
                           self._ws_handler,
                           name='websockets')
        app.on_shutdown.append(self.shutdown)

    async def send_to_all(self, msg: dict) -> None:
        _log.debug("send_to_all: enterinmg function, _ws_list: %s",
                   self._ws_list)
        for ws in self._ws_list:
            _log.debug("send_to_all: in for loop, ws: %s", ws)
            await ws.send_str(msg)
        return None

    def subscribe(self, scope: str, callback: callable) -> None:
        """Install a callback function for given scope.

        :param scope: name of scope that asks for websocket
        :type scope: str
        :param callback: callback function that is called when data arrives
        :type callback: function with parameters msg
        :return: None
        :rtype: None
        """
        self._subscriptions[scope] = callback
        return None

    async def _ws_handler(self, request) -> web.WebSocketResponse:
        max_msg_size = self._options.websockets_max_receive_size
        ws = web.WebSocketResponse(max_msg_size=max_msg_size)
        await ws.prepare(request)

        self._ws_list.append(ws)

        try:
            async for msg in ws:
                _log.debug(msg)
                if msg.type == web.WSMsgType.TEXT:
                    await self._dispatch_message(msg.data)
                else:
                    _log.error(f'Unknown message type: {msg}')
                    await asyncio.sleep(3)
                    break
        except Exception as e:
            _log.error(f'Exeption while reading from websocket: {e}')
            await asyncio.sleep(3)
        finally:
            _log.debug(f'Removing closed websocket: {ws}')
            self._ws_list.remove(ws)
        return ws

    async def _dispatch_message(self, msg: dict) -> bool:
        msg = json.loads(msg)
        scope = msg.get('scope', '')
        if scope == "authenticate_v1":
            # TODO Implemet a one-time authentication
            return True

        if len(scope) == 0:
            _log.error(f'Message not for a scope: {msg}')
            await asyncio.sleep(3)
            return False

        token = msg.get('token', '')
        sequence = msg.get('sequence', '')

        if len(sequence) == 0 or len(token) == 0:
            _log.error(f"Websocket-Auth: missing var {sequence},{token}")
            await asyncio.sleep(3)
            return False

        cursor = self._db.conn.cursor()
        cursor.execute('''SELECT * FROM users_tokens
                    WHERE sequence = ?''', (sequence,))
        item = cursor.fetchone()
        if item is None:
            _log.error(f"Websocket-Auth: unknown {sequence},{token}")
            await asyncio.sleep(3)
            return False

        token = token.encode('utf-8')
        if bcrypt.checkpw(token, item['hash']):
            _log.debug(f'authenticate: {token}, sequence: {sequence}')
        else:
            _log.error(f'no authenticate: {token}, sequence: {sequence}')
            await asyncio.sleep(3)
            return False

        callback_func = self._subscriptions.get(scope, None)
        if callback_func is None:
            _log.error(f'Message for an unknown scope: {msg}')
            await asyncio.sleep(3)
            return False
        try:
            await callback_func(msg)
        except Exception as e:
            _log.error(f'Exeption in {callback_func}: {e}')
            await asyncio.sleep(3)
            return False
        return True

    async def shutdown(self, app) -> None:
        loop = asyncio.get_event_loop()
        for ws in self._ws_list:
            _log.debug(f'calling ws.close for: {ws}')
            loop.create_task(ws.close())
        return None



import asyncio
import configargparse
import pkg_resources
import sys
from pathlib import Path
import socket
import secrets

import logging
from logging.handlers import RotatingFileHandler
from logging.handlers import TimedRotatingFileHandler

from riego.db import setup_db
from riego.mqtt import setup_mqtt
from riego.cloud import setup_cloud
from riego.boxes import setup_boxes
from riego.valves import setup_valves
from riego.timer import setup_timer
from riego.model.parameters import setup_parameters


from riego.web.websockets import setup_websockets
from riego.web.routes import setup_routes
from riego.web.error_pages import setup_error_pages
from riego.web.security import current_user_ctx_processor


from aiohttp import web
import jinja2
import aiohttp_jinja2
import aiohttp_debugtoolbar
from aiohttp_remotes import setup as setup_remotes, XForwardedRelaxed

from aiohttp_session import setup as session_setup
from aiohttp_session.memcached_storage import MemcachedStorage
import aiomcache

from riego import __version__

PRIMARY_INI_FILE = 'riego.conf'


async def on_startup(app):
    logging.getLogger(__name__).debug("on_startup")


async def on_shutdown(app):
    logging.getLogger(__name__).debug("on_shutdown")


async def on_cleanup(app):
    logging.getLogger(__name__).debug("on_cleanup")


def main():
    options = _get_options()

    _setup_logging(options=options)

    if sys.version_info >= (3, 8) and options.WindowsSelectorEventLoopPolicy:
        asyncio.DefaultEventLoopPolicy = asyncio.WindowsSelectorEventLoopPolicy  # noqa: E501

    if sys.platform != "win32":
        import uvloop  # pylint: disable=import-error
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    web.run_app(run_app(options=options),
                host=options.http_server_bind_address,
                port=options.http_server_bind_port,
                # access_log=_setup_access_log(options=options)
                )


async def run_app(options=None):
    loop = asyncio.get_event_loop()

    if options.enable_asyncio_debug_log:
        loop.set_debug(True)

    app = web.Application()

    app.on_startup.append(on_startup)
    app.on_shutdown.append(on_shutdown)
    app.on_cleanup.append(on_cleanup)

    app['version'] = __version__
    app['options'] = options

    db = setup_db(options=options)
    websockets = setup_websockets(app=app, db=db, options=options)
    parameters = setup_parameters(app=app, db=db, options=options)
    mqtt = setup_mqtt(app=app, options=options)
    setup_boxes(options=options, db=db, mqtt=mqtt)
    valves = setup_valves(options=options, db=db,
                          mqtt=mqtt, websockets=websockets)
    setup_timer(app=app, options=options, db=db,
                mqtt=mqtt, valves=valves, parameters=parameters)
    setup_cloud(app=app, parameters=parameters, options=options)

    mcache = aiomcache.Client(options.memcached_host, options.memcached_port)
    session_setup(app, MemcachedStorage(mcache))

    async def mcache_shutdown(app):
        await mcache.close()
    app.on_shutdown.append(mcache_shutdown)

    loader = jinja2.FileSystemLoader(options.http_server_template_dir)
    aiohttp_jinja2.setup(app,
                         loader=loader,
                         # enable_async=True,
                         context_processors=[current_user_ctx_processor],
                         )

    await setup_remotes(app, XForwardedRelaxed())
    setup_routes(app, options=options)

    if not options.verbose:
        setup_error_pages(app)

    if options.enable_aiohttp_debug_toolbar:
        aiohttp_debugtoolbar.setup(
            app, check_host=False, intercept_redirects=False)

# Put app as subapp under main_app and create an approbiate redirection
    main_app = web.Application()

    if parameters.cloud_identifier is None:
        parameters.cloud_identifier = secrets.token_urlsafe(12)

    async def main_app_handler(request):
        raise web.HTTPSeeOther(f'/{parameters.cloud_identifier}/')

    main_app.router.add_get('/', main_app_handler)
    main_app.add_subapp(f'/{parameters.cloud_identifier}/', app)

    logging.getLogger(__name__).info(f'Start {parameters.cloud_identifier}')
    return main_app


def _setup_access_log(options=None):
    formatter = logging.Formatter("%(message)s")
    Path(options.http_access_log_file).parent.mkdir(
        parents=True, exist_ok=True)
    file_handler = RotatingFileHandler(
        options.http_access_log_file, mode='a',
        maxBytes=options.http_access_log_max_bytes,
        backupCount=options.http_access_log_backup_count,
        encoding=None, delay=0)
    file_handler.setFormatter(formatter)
    file_handler.setLevel(logging.DEBUG)

    access_log = logging.getLogger("aiohttp.access")
    access_log.setLevel(logging.DEBUG)
    access_log.addHandler(file_handler)
    return access_log


def _setup_logging(options=None):
    formatter = logging.Formatter(
        "%(asctime)s;%(levelname)s;%(name)s;%(message)s")
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    if options.verbose:
        level = logging.DEBUG
    else:
        level = logging.INFO
    Path(options.log_file).parent.mkdir(parents=True, exist_ok=True)
    file_handler = TimedRotatingFileHandler(
        options.log_file,
        when='midnight',
        backupCount=options.log_backup_count)
    file_handler.setFormatter(formatter)
    if options.log_backup_count == -1:
        logging.basicConfig(level=level, handlers=[stream_handler])
    else:
        logging.basicConfig(level=level,
                            handlers=[stream_handler, file_handler])

    if options.enable_gmqtt_debug_log:
        logging.getLogger("gmqtt").setLevel(logging.DEBUG)
    else:
        logging.getLogger("gmqtt").setLevel(logging.ERROR)

    if options.enable_ssh_debug_log:
        logging.getLogger("asyncssh").setLevel(logging.DEBUG)
    else:
        logging.getLogger("asyncssh").setLevel(logging.ERROR)

    if options.enable_aiohttp_access_log:
        logging.getLogger("aiohttp.access").setLevel(logging.DEBUG)
    else:
        logging.getLogger("aiohttp.access").setLevel(logging.ERROR)


def _get_options():
    p = configargparse.ArgParser(
        default_config_files=['/etc/riego/conf.d/*.conf',
                              '~/.riego.conf',
                              PRIMARY_INI_FILE],
        args_for_writing_out_config_file=['-w',
                                          '--write-out-config-file'])
    p.add('-c', '--config', is_config_file=True, env_var='RIEGO_CONF',
          required=False, help='config file path')
# Database
    p.add('-d', '--db_filename', help='Path and name for DB file',
          default='db/riego.db')
    p.add('--db_migrations_dir',
          help='path to database migrations directory',
          default=pkg_resources.resource_filename('riego', 'migrations'))
# Logging
    p.add('-l', '--log_file', help='Full path to logfile',
          default='log/riego.log')
    p.add('--log_backup_count', help='How many files to rotate, -1 to disable',
          default=20, type=int)
# Secrets & Scurity
    p.add('--max_age_remember_me', type=int, default=7776000)
    p.add('--cookie_name_remember_me', default="remember_me")
    p.add('--reset_admin', help='Reset admin-pw to given value an exit')
    p.add('--websockets_max_receive_size', type=int, default=1024)
    p.add('--ssh_key_algorithm', default='ssh-ed25519')
# Session keys
    p.add('--session_key_websocket_auth', default='websocket_auth')
    # TODO
    p.add('--session_key_user_id', default='user_id',
          help='# TODO not used Yet')
# HTTP-Server
    p.add('--http_server_bind_address',
          help='http-server bind address', default='0.0.0.0')
    p.add('--http_server_bind_port', help='http-server bind port',
          default=8080, type=int)
    p.add('--http_server_static_dir',
          help='Serve static html files from this directory',
          default=pkg_resources.resource_filename('riego.web', 'static'))
    p.add('--http_server_template_dir',
          help='Serve template files from this directory',
          default=pkg_resources.resource_filename('riego.web', 'templates'))
    p.add('--websocket_path', help='url path for websocket',
          default="/ws")
    p.add('--http_access_log_file', help='Full path to access logfile',
          default='log/access.log')
    p.add('--http_access_log_max_bytes', help='Maximum file size in bytes',
          default=1024*300, type=int)
    p.add('--http_access_log_backup_count', help='How many files to rotate',
          default=3, type=int)
# Memcache
    p.add('--memcached_host', help='IP adress of memcached host',
          default='127.0.0.1')
    p.add('--memcached_port', help='Port of memcached service',
          default=11211, type=int)
# Cloud-Server
#    p.add('--cloud_identifier', help='Unique id for Cloud Identity')
    p.add('--cloud_api_url', help='Hostname for cloud service',
          default='https://api.riego.cloud/api_20210221/')
    p.add('--ssh_known_hosts', help='SSH-Server verification',
          default=pkg_resources.resource_filename('riego', 'ssh/known_hosts'))

# Directories
    p.add('--base_dir', help='Change only if you know what you are doing',
          default=Path(__file__).parent)
# MQTT-Broker
    p.add('-m', '--mqtt_broker_host', help='IP adress of MQTT-Broker Host',
          default='127.0.0.1')
    p.add('-p', '--mqtt_broker_port', help='Port of MQTT-Broker Host',
          default=1883, type=int)
    p.add('--mqtt_client_id', help='Client ID for MQTT-Connection',
          default=f'riego_ctrl_{socket.gethostname()}')

    p.add('--mqtt_result_subscription', help='used by class valves',
          default="stat/+/RESULT")
    p.add('--mqtt_lwt_subscription', help='used by class boxes',
          default="tele/+/LWT")
    p.add('--mqtt_state_subscription', help='used by class boxes',
          default="tele/+/STATE")
    p.add('--mqtt_info1_subscription', help='used by class boxes',
          default="tele/+/INFO1")
    p.add('--mqtt_info2_subscription', help='used by class boxes',
          default="tele/+/INFO2")

    p.add('--mqtt_sensor_subscription', help='yet not used',
          default="tele/+/SENSOR")

    p.add('--mqtt_cmnd_prefix', help='',
          default="cmnd")
    p.add('--mqtt_keyword_ON', help='',
          default="ON")
    p.add('--mqtt_keyword_OFF', help='',
          default="OFF")
# Parameter Default Values
    p.add('--parameters_smtp_hostname', default="smtp.gmail.com")
    p.add('--parameters_smtp_port', type=int, default=465)
    p.add('--parameters_smtp_security', default="use_tls=True")
    p.add('--parameters_smtp_user', default="user")
    p.add('--parameters_smtp_password', default="password")
    p.add('--parameters_smtp_from', default="riego@localhost")

    p.add('--parameters_start_time_1', default="19:00")
    p.add('--parameters_max_duartion', default="240")
# Debug
    p.add('--enable_aiohttp_debug_toolbar', action='store_true')
    p.add('--enable_aiohttp_access_log', action='store_true')
    p.add('--enable_asyncio_debug_log', action='store_true')
    p.add('--enable_ssh_debug_log', action='store_true')
    p.add('--enable_gmqtt_debug_log', action='store_true')
    p.add('--enable_timer_dev_mode', action='store_true')
    p.add('--WindowsSelectorEventLoopPolicy', action='store_true')

# Version, Help, Verbosity
    p.add('-v', '--verbose', help='verbose', action='store_true')
    p.add('--version', help='Print version and exit', action='store_true')
    p.add('--defaults', help='Print options with default values and exit',
          action='store_true')

    options = p.parse_args()
    if options.verbose:
        print(p.format_values())

    try:
        with open(PRIMARY_INI_FILE, 'xt') as f:
            for item in vars(options):
                f.write(f'# {item}={getattr(options, item)}\n')
    except IOError:
        pass

    if options.defaults:
        for item in vars(options):
            print(f'# {item}={getattr(options, item)}')
        exit(0)

    if options.version:
        print('Version: ', __version__)
        exit(0)

    if options.reset_admin:
        _reset_admin(options)
        exit(0)

    return options


def _reset_admin(options):
    from riego.db import setup_db
    import bcrypt

    db = setup_db(options=options)
    password = options.reset_admin

    if len(password) > 0:
        password = password.encode('utf-8')
        password = bcrypt.hashpw(password, bcrypt.gensalt())
        password = password.decode('utf-8')
        cursor = db.conn.cursor()
        cursor.execute('''UPDATE users
                        SET password = ?
                        WHERE id = ? ''', (password, 1))
        db.conn.commit()
        if cursor.rowcount == 1:
            print(f'Succesfully reset Admin PW: {password}')
        else:
            print('Unable to reset Admin PW:')
    db.conn.close()



import aiohttp_jinja2
from aiohttp import web

from riego.db import get_db
from riego.web.security import raise_permission


router = web.RouteTableDef()


def setup_routes_events(app):
    app.add_routes(router)


@router.get("/events", name='events')
@aiohttp_jinja2.template('events/index.html')
async def event_index(request):
    await raise_permission(request, permission=None)
    cursor = get_db().conn.cursor()
    cursor.execute('''SELECT events.*,
                valves.name AS valves_name,
                valves.id AS valves_id,
                valves.duration AS valves_duration
                FROM events, valves
                WHERE events.valve_id = valves.id
                ORDER BY events.created_at DESC''')
    items = cursor.fetchall()
    get_db().conn.commit()
    return {'items': items}


@router.get("/events/{item_id}/filter", name='events_item_filter')
@aiohttp_jinja2.template('events/index.html')
async def event_filter(request):
    await raise_permission(request, permission=None)
    item_id = request.match_info["item_id"]
    cursor = get_db().conn.cursor()
    cursor.execute('''SELECT events.*, valves.name
                FROM events, valves
                WHERE events.valve_id = valves.id
                AND valves.id = ?
                ORDER BY events.created_at DESC''', (item_id,))
    items = cursor.fetchall()
    get_db().conn.commit()
    return {'items': items}



from typing import Any, Dict
import aiohttp_jinja2
from aiohttp import web

from sqlite3 import IntegrityError
from riego.db import get_db
from riego.web.security import raise_permission

from logging import getLogger
_log = getLogger(__name__)

router = web.RouteTableDef()


def setup_routes_valves(app):
    app.add_routes(router)


@router.get("/valves", name='valves')
@aiohttp_jinja2.template("valves/index.html")
async def index(request: web.Request) -> Dict[str, Any]:
    await raise_permission(request, permission="")
    cursor = get_db().conn.cursor()
    cursor.execute('''SELECT valves.*,
                        boxes.name AS box_name,
                        boxes.topic AS box_topic,
                        boxes.id AS box_id
                        FROM valves, boxes
                        WHERE valves.box_id = boxes.id''')
    items = cursor.fetchall()
    get_db().conn.commit()
    return {"items": items}


@router.get("/valves/new", name='valves_new')
@aiohttp_jinja2.template("valves/new.html")
async def new(request: web.Request) -> Dict[str, Any]:
    await raise_permission(request, permission="")
    return {}


@router.post("/valves/new")
async def new_apply(request: web.Request) -> Dict[str, Any]:
    await raise_permission(request, permission="")
    item = await request.post()
    try:
        with get_db().conn:
            cursor = get_db().conn.execute(
                ''' INSERT INTO valves
                (name, channel_nr, box_id)
                VALUES (?, ?, ?) ''',
                (item['name'], item['channel_nr'], item['box_id']))
    except IntegrityError as e:
        _log.debug(f'box.view add: {e}')
        raise web.HTTPSeeOther(request.app.router['valves_new'].url_for())
    else:
        item_id = str(cursor.lastrowid)
        raise web.HTTPSeeOther(
            request.app.router['valves_item_view'].url_for(item_id=item_id))
    return {}  # not reached


@router.get("/valves/{item_id}", name='valves_item_view')
@aiohttp_jinja2.template("valves/view.html")
async def view(request: web.Request) -> Dict[str, Any]:
    await raise_permission(request, permission="")
    item_id = request.match_info["item_id"]
    cursor = get_db().conn.cursor()
    cursor.execute('''SELECT valves.*,
                boxes.name AS box_name,
                boxes.topic AS box_topic,
                boxes.id AS box_id
                FROM valves, boxes
                WHERE valves.box_id = boxes.id AND valves.id=?''', (item_id,))
    item = cursor.fetchone()
    get_db().conn.commit()
    if item is None:
        raise web.HTTPSeeOther(request.app.router['valves'].url_for())
    return {"item": item}


@router.get("/valves/{item_id}/edit", name='valves_item_edit')
@aiohttp_jinja2.template("valves/edit.html")
async def edit(request: web.Request) -> Dict[str, Any]:
    await raise_permission(request, permission="")
    item_id = request.match_info["item_id"]
    cursor = get_db().conn.cursor()
    cursor.execute('''SELECT valves.*,
                boxes.name AS box_name,
                boxes.topic AS box_topic
                FROM valves, boxes
                WHERE valves.box_id = boxes.id AND valves.id=?''', (item_id,))
    item = cursor.fetchone()
    get_db().conn.commit()
    if item is None:
        raise web.HTTPSeeOther(request.app.router['valves'].url_for())
    return {"item": item}


@router.post("/valves/{item_id}/edit")
async def edit_apply(request: web.Request) -> web.Response:
    await raise_permission(request, permission="")
    item_id = request.match_info["item_id"]
    item = await request.post()
    try:
        with get_db().conn:
            get_db().conn.execute(
                ''' UPDATE valves
                    SET name = ?, remark = ?,
                    duration = ?, interval = ?,
                    is_running = ? , is_enabled= ?,
                    is_hidden = ?,  prio = ?
                    WHERE id = ? ''',
                (item['name'], item['remark'],
                 item['duration'], item['interval'],
                 item['is_running'], item['is_enabled'],
                 item['is_hidden'], item['prio'], item_id))
    except IntegrityError as e:
        _log.debug(f'box.view edit: {e}')
        raise web.HTTPSeeOther(
            request.app.router['valves_item_edit'].url_for(item_id=item_id))
    else:
        raise web.HTTPSeeOther(
            request.app.router['valves_item_view'].url_for(item_id=item_id))
    return {}  # Not reached


@router.get("/valves/{item_id}/delete", name='valves_item_delete')
async def delete(request: web.Request) -> web.Response:
    await raise_permission(request, permission="")
    item_id = request.match_info["item_id"]
    try:
        with get_db().conn:
            get_db().conn.execute(
                'DELETE FROM valves WHERE id = ?',
                (item_id,))
    except IntegrityError as e:
        _log.debug(f'valves delete: {e}')
    raise web.HTTPSeeOther(request.app.router['valves'].url_for())
    return {}  # Not reached



import aiohttp_jinja2
from aiohttp import web


async def handle_401(request):
    return aiohttp_jinja2.render_template('401.html', request, {}, status=401)


async def handle_403(request):
    return aiohttp_jinja2.render_template('403.html', request, {}, status=403)


async def handle_404(request):
    return aiohttp_jinja2.render_template('404.html', request, {}, status=404)


async def handle_500(request):
    return aiohttp_jinja2.render_template('500.html', request, {}, status=500)


def create_error_middleware(overrides):

    @web.middleware
    async def error_middleware(request, handler):
        try:
            return await handler(request)
        except web.HTTPException as ex:
            override = overrides.get(ex.status)
            if override:
                return await override(request)

            raise
        except Exception:
            return await overrides[500](request)

    return error_middleware


def setup_error_pages(app=None):
    error_middleware = create_error_middleware({
        401: handle_401,
        403: handle_403,
        404: handle_404,
        500: handle_500
    })
    app.middlewares.append(error_middleware)
